package cz.fidentis.analyst.visitors.mesh;

import cz.fidentis.analyst.mesh.MeshVisitor;
import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import cz.fidentis.analyst.mesh.core.MeshTriangle;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.vecmath.Vector3d;

/**
 * Abstract class for algorithms calculating curvatures of mesh vertices.
 * @see https://computergraphics.stackexchange.com/questions/1718/what-is-the-simplest-way-to-compute-principal-curvature-for-a-mesh-triangle
 * <p>
 * <b>All curvature algorithms suppose that the triangle vertices are oriented clockwise!</b>
 * </p>
 * 
 * TODO: FeaturePoints/src/cz/fidentis/featurepoints/curvature/Curvature_jv.java
 * TODO: http://rodolphe-vaillant.fr/?e=20
 * 
 * @author Natalia Bebjakova
 * @author Radek Oslejsek
 */
public abstract class Curvature extends MeshVisitor {
    
    private final Map<MeshFacet, List<Double>> curvatures = new HashMap<>();
    
    /**
     * Returns curvatures for all inspected mesh facets. The order corresponds to
     * the order of vertices, i.e., i-th value represents the curvature of i-th mesh vertex.
     * 
     * @return curvatures of inspected mesh facets.
     */
    public Map<MeshFacet, List<Double>> getCurvatures() {
        return Collections.unmodifiableMap(curvatures);
    }
    
    @Override
    public void visitMeshFacet(final MeshFacet facet) {
        synchronized (this) {
            if (curvatures.containsKey(facet)) {
                return; // already visited facet
            }
            curvatures.put(facet, new ArrayList<>());
            if (!facet.hasVertexNormals()) {
                facet.calculateVertexNormals();
            }
            //facet.calculateVoronoiPoints();
        }
        
        //final List<MeshTriangle> triangles = facet.getTriangles();
        final List<CurvTriangle> triangles = precomputeTriangles(facet);
        for (int i = 0; i < facet.getNumberOfVertices(); i++) {
            curvatures.get(facet).add(calculateCurvature(facet, triangles, i));
        }
    }
    
    /**
     * 
     * @param facet
     * @param triangles
     * @param vertA The "central point" for the 1-ring neighborhood
     * @return 
     */
    protected abstract double calculateCurvature(MeshFacet facet, List<CurvTriangle> triangles, int vertA);
    
    protected List<CurvTriangle> precomputeTriangles(MeshFacet facet) {
        List<CurvTriangle> ret = new ArrayList<>(facet.getNumTriangles());
        for (MeshTriangle tri: facet) {
            ret.add(new CurvTriangle(facet, tri));
        }
        return ret;
    }
    
    protected double computeArea(CurvTriangle ctri, MeshPoint vertA) {
        double alpha = ctri.alpha(vertA);
        double beta = ctri.beta(vertA);
        double gamma = ctri.gamma(vertA);
            
        double piHalf = Math.PI / 2.0; // 90 degrees
        if (alpha >= piHalf || beta >= piHalf || gamma >= piHalf) { // check for obtuse angle
            return (alpha > piHalf) ? ctri.area() / 2.0 : ctri.area() / 4.0;
        } else {
            double cotBeta = ctri.betaCotan(vertA);
            double cotGamma = ctri.gammaCotan(vertA);
            double ab = ctri.lengthSquaredAB(vertA);
            double ac = ctri.lengthSquaredAC(vertA);
            return (ab * cotGamma + ac * cotBeta) / 8.0;
        }
    }
    
    /**
     * Helper class that caches values used multiples times in triangle during the curvature computation.
     * <ul>
     * <li>A = central point of the 1-ring neighborhood</li>
     * <li>B = point "on the left" (previous point in the clockwise orientation of the triangle</li>
     * <li>C = point "on the right" (next point in the clockwise orientation of the triangle</li>
     * </ul>
     * 
     * @author Radek Oslejsek
     */
    protected class CurvTriangle {
        
        private MeshFacet facet;
        
        public MeshTriangle tri;
        
        public double v1Angle;
        public double v2Angle;
        public double v3Angle;
        
        public double v1AngleCotang;
        public double v2AngleCotang;
        public double v3AngleCotang;
        
        public double v12lengthSquared;
        public double v23lengthSquared;
        public double v31lengthSquared;
        
        public CurvTriangle(MeshFacet facet, MeshTriangle tri) {
            this.tri = tri;
            
            Vector3d a = new Vector3d(facet.getVertex(tri.index3).getPosition());
            a.sub(facet.getVertex(tri.index2).getPosition());
            
            Vector3d b = new Vector3d(facet.getVertex(tri.index1).getPosition());
            b.sub(facet.getVertex(tri.index3).getPosition());
            
            Vector3d c = new Vector3d(facet.getVertex(tri.index2).getPosition());
            c.sub(facet.getVertex(tri.index1).getPosition());
            
            this.v23lengthSquared = a.lengthSquared();
            this.v31lengthSquared = b.lengthSquared();
            this.v12lengthSquared = c.lengthSquared();
            
            a.normalize();
            b.normalize();
            c.normalize();
            
            b.scale(-1.0);
            double cos1 = c.dot(b);
            
            c.scale(-1.0);
            double cos2 = a.dot(c);
            
            a.scale(-1.0);
            b.scale(-1.0);
            double cos3 = a.dot(b);
            
            this.v1Angle = Math.acos(cos1);
            this.v2Angle = Math.acos(cos2);
            this.v3Angle = Math.acos(cos3);
            
            this.v1AngleCotang = 1.0 / Math.tan(v1Angle);
            this.v2AngleCotang = 1.0 / Math.tan(v2Angle);
            this.v3AngleCotang = 1.0 / Math.tan(v3Angle);
        }
        
        public double area() {
            return 0.5 * v12lengthSquared * v31lengthSquared * Math.sin(v1Angle);
        }
        
        /**
         * Returns vertex A. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return vertex A.
         */
        public Vector3d vertA(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return tri.vertex1.getPosition();
            } else if (facetPoint == tri.vertex2) {
                return tri.vertex2.getPosition();
            } else if (facetPoint == tri.vertex3) {
                return tri.vertex3.getPosition();
            }
            return null;
        }

        /**
         * Returns vertex B. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return vertex B.
         */
        public Vector3d vertB(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return tri.vertex3.getPosition();
            } else if (facetPoint == tri.vertex2) {
                return tri.vertex1.getPosition();
            } else if (facetPoint == tri.vertex3) {
                return tri.vertex2.getPosition();
            }
            return null;
        }

        /**
         * Returns vertex C. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return vertex C.
         */
        public Vector3d vertC(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return tri.vertex2.getPosition();
            } else if (facetPoint == tri.vertex2) {
                return tri.vertex3.getPosition();
            } else if (facetPoint == tri.vertex3) {
                return tri.vertex1.getPosition();
            }
            return null;
        }

        /**
         * Returns cached angle in the vertex A. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Alpha angle.
         */
        public double alpha(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v1Angle;
            } else if (facetPoint == tri.vertex2) {
                return v2Angle;
            } else if (facetPoint == tri.vertex3) {
                return v3Angle;
            }
            return Double.NaN;
        }

        /**
         * Returns cached angle in the vertex B.
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Beta angle.
         */
        public double beta(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v3Angle;
            } else if (facetPoint == tri.vertex2) {
                return v1Angle;
            } else if (facetPoint == tri.vertex3) {
                return v2Angle;
            }
            return Double.NaN;
        }

        /**
         * Returns cached angle in the vertex C. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Gamma angle.
         */
        public double gamma(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v2Angle;
            } else if (facetPoint == tri.vertex2) {
                return v3Angle;
            } else if (facetPoint == tri.vertex3) {
                return v1Angle;
            }
            return Double.NaN;
        }

        /**
         * Returns cached cotangent of the angle in the vertex A. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Cotangent of the alpha angle.
         */
        public double alphaCotan(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v1AngleCotang;
            } else if (facetPoint == tri.vertex2) {
                return v2AngleCotang;
            } else if (facetPoint == tri.vertex3) {
                return v3AngleCotang;
            }
            return Double.NaN;
        }

        /**
         * Returns cached cotangent of the angle in the vertex B.
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Cotangent of the beta angle.
         */
        public double betaCotan(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v3AngleCotang;
            } else if (facetPoint == tri.vertex2) {
                return v1AngleCotang;
            } else if (facetPoint == tri.vertex3) {
                return v2AngleCotang;
            }
            return Double.NaN;
        }

        /**
         * Returns cached cotangent of the angle in the vertex C. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Cotangent of the gamma angle.
         */
        public double gammaCotan(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v2AngleCotang;
            } else if (facetPoint == tri.vertex2) {
                return v3AngleCotang;
            } else if (facetPoint == tri.vertex3) {
                return v1AngleCotang;
            }
            return Double.NaN;
        }
        
        /**
         * Returns cached squared length of the edge AB (opposite to the vertex A). 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Squared length of the edge AB.
         */
        public double lengthSquaredAB(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v31lengthSquared;
            } else if (facetPoint == tri.vertex2) {
                return v12lengthSquared;
            } else if (facetPoint == tri.vertex3) {
                return v23lengthSquared;
            }
            return Double.NaN;
        }
        
        /**
         * Returns cached squared length of the edge AC (opposite to the vertex B). 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Squared length of the edge AC.
         */
        public double lengthSquaredAC(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v12lengthSquared;
            } else if (facetPoint == tri.vertex2) {
                return v23lengthSquared;
            } else if (facetPoint == tri.vertex3) {
                return v31lengthSquared;
            }
            return Double.NaN;
        }
        
        /**
         * Returns cached squared length of the edge BC (opposite to the vertex A). 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Squared length of the edge BC.
         */
        public double lengthSquaredBC(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v23lengthSquared;
            } else if (facetPoint == tri.vertex2) {
                return v31lengthSquared;
            } else if (facetPoint == tri.vertex3) {
                return v12lengthSquared;
            }
            return Double.NaN;
        }
    }

}
