package cz.fidentis.analyst.visitors.mesh;

import cz.fidentis.analyst.mesh.core.MeshFacet;
import java.util.List;
import javax.vecmath.Vector3d;

/**
 * Mean curvature based on solving Laplace-Beltrami mean operator by cotangent discretization.
 * 
 * @see https://computergraphics.stackexchange.com/questions/1718/what-is-the-simplest-way-to-compute-principal-curvature-for-a-mesh-triangle
 * @see http://rodolphe-vaillant.fr/?e=20
 * 
 * @author Natalia Bebjakova
 * @author Radek Oslejsek
 */
public class MeanCurvature extends Curvature {
    
    @Override
    protected double calculateCurvature(MeshFacet facet, List<CurvTriangle> triangles, int vertA) {
        List<Integer> neighbouringTriangles = facet.getCornerTable().getTriangleIndexesByVertexIndex(vertA);
        
        if (neighbouringTriangles.isEmpty()) {
            return Double.NaN;
        }
        
        double sumArea = 0.0;
        Vector3d pointSum = new Vector3d();
        
        // for all surrounding triangles:
        for (int i = 0; i < neighbouringTriangles.size(); i++) {
            CurvTriangle ctri = triangles.get(neighbouringTriangles.get(i));
            CurvTriangle tNext = triangles.get(neighbouringTriangles.get((i + 1) % neighbouringTriangles.size()));
            
            sumArea += computeArea(ctri, facet.getVertex(vertA));
            
            Vector3d aux = new Vector3d(ctri.vertC(facet.getVertex(vertA)));
            aux.sub(ctri.vertA(facet.getVertex(vertA)));
            aux.scale(ctri.betaCotan(facet.getVertex(vertA)) + tNext.gammaCotan(facet.getVertex(vertA)));
            pointSum.add(aux);
        }
        
        return 0.25 * sumArea * pointSum.length();
            
    }
}
